#!/usr/local/bin/perl -w
#
# Fixes structural problems:
#   - when list occurs inside paragraph, split paragraph around list
#   - unwrap <literal> if occurs in <screen> or <programlisting>
#   - move <indexterm>s out of titles.
#
# Works by parsing XML document to build an object tree,
# processing nodes in order,
# and serializing the tree back to XML.
#
# Usage: dbfix <top-xml-file>
#

use strict;

# -------------------------------------------------------------------------
# GLOBAL DATA
# -------------------------------------------------------------------------

# XML Object Tree Structure:
#
# node --> type
#      --> name
#      --> data
#      --> parent
#      --> [children]
#
# where type is one of: element, attribute, 
# PI, declaration, comment, cdms, text, or root; 
# name further specifies the node variant;
# data is the content of the node;
# and [children] is a list of nodes.
#

my %unwrap_element =    # nodes to unwrap: context => [node]
                             (
                              'screen' => ['literal'],
                              'programlisting' => ['literal'],
			      );
my %raise_and_move_backward = # nodes to raise in front: context => [node]
                             (
			      'title' => [ 'indexterm', 'icon' ],
			      'refname' => [ 'indexterm', 'icon' ],
			      'refentrytitle' => [ 'indexterm', 'icon' ]
			      );
my %raise_in_place =    # nodes to raise in place: context => [node]
                             (
			      'para' => [
					 'variablelist',
					 'orderedlist',
					 'itemizedlist',
					 'simplelist',
					 ],
			      );
my %frefs;              # file entities declared in internal subset
my $rootnode;           # root of XML object tree


# get the top-level file for processing
my $file = shift @ARGV;
if( $file && -e $file ) {
    &main();
    exit(0);
} else {
    print STDERR "\nUsage: $0 <top-xml-file>\n\n";
    exit(1);
}


# -------------------------------------------------------------------------
# SUBROUTINES
# -------------------------------------------------------------------------

# main
# ----
# Top level subroutine.
# 
sub main {
    # 1. Input XML files, construct text buffer of document entity
    my $text = &process_file( $file );
    
    # 2. Build object tree of XML nodes
    $rootnode = &make_node( 'root', undef, undef, &parse_document( $text ));

    # 3. Process nodes, adding number atts, etc.
    &process_tree( $rootnode );

    # 4. Translate the tree back into text
    $text = &serialize_tree( $rootnode );

    # 5. Output text to separate files again
    &output_text( $file, $text );
}


# process_file
# ------------
# Get text from all XML files in document. 
#
sub process_file {
    my( $file ) = @_;
    unless( open( F, $file )) {
	print STDERR "Can't open \"$file\" for reading.\n";
	return "";
    }
    my @lines = <F>;
    close F;
    my $line;
    my $buf = "";
    foreach $line (@lines) {

	# Replace external entity references with file contents
	if( $line =~ /\&([^;]+);/ && $frefs{$1} ) {
	    my( $pre, $ent, $post ) = ($`, $&, $' );
	    my $newfile = $frefs{$1};
	    $buf .= $pre . $ent . "\n<?xml-file startfile: $newfile ?>" .
		&process_file( $frefs{$1} ) . "<?xml-file endfile ?>" . 
		    $post;
	} else {
	    $buf .= $line;
	}

	# Add declared external entities to the list.
	# NOTE: we do not handle PUBLIC identifiers!
	$frefs{ $1 } = $2 
	    if( $line =~ /<!ENTITY\s+(\S+)\s+SYSTEM\s+\"([^\"]+)/ );
    }
    return $buf;
}


# parse_document
# --------------
# Read nodes at top level of document.
#
sub parse_document {
    my( $text ) = @_;
    my @nodes = ();
    while( $text ) {
	my $node;
	( $node, $text ) = &get_node( $text );
	push( @nodes, $node );
	sleep(1);
    }
    return @nodes;
}


# get_node
# --------
# Given a piece of XML text, return the first node found
# and the rest of the text string.
#
sub get_node {
    my( $text ) = @_;
    my $node;
    my( $type, $name, $data ) = ('','','');
    my @children = ();

    # text
    if( $text =~ /^[^<]+/ ) {
	$text = $';
	$type = 'text';
	$data = $&;

    # comment, marked section, declaration
    } elsif( $text =~ /^\s*<\!/ ) {
	if( $text =~ /^\s*<\!--(.*?)-->/s ) {
	    $text = $';
	    $type = 'comment';
	    $data = $1;
	    if( $data =~ /--/ ) {
		&parse_error( "comment contains partial delimiter (--)" );
	    }
	} elsif( $text =~ /^\s*<\!\[\s*CDATA\s*\[/ ) {
	    $text = $';
	    $type = 'CDMS';
	    if( $text =~ /\]\]>/ ) {
		$text = $';
		$data = $`;
	    } else {
		&parse_error( "CDMS" );
	    }
	} elsif( $text =~ /^\s*<!DOCTYPE(.*?\])>\s*/s ) {
	    $text = $';
	    $name = 'DOCTYPE';
	    $type = 'declaration';
	    $data = $1;
	} else {
	    &parse_error( "declaration syntax" );
	}

    # processing instruction
    } elsif( $text =~ /^\s*<\?/ ) {
	if( $text =~ /^\s*<\?\s*([^\s\?]+)\s*(.*?)\s*\?>\s*/s ) {
	    $text = $';
	    $name = $1;
	    $type = 'PI';
	    $data = $2;
	} else {
	    &parse_error( "PI syntax" );
	}

    # element
    } elsif( $text =~ /\s*</ ) {

	# empty element with atts
	if( $text =~ /^\s*<([^\/\s>]+)\s+([^\s>][^>]+)\/>/) {
	    $text = $';
	    $name = $1;
	    $type = 'empty-element';
	    my $atts = $2;
	    push( @children, &parse_atts( $atts ));

	# empty element, no atts
	} elsif( $text =~ /^\s*<([^\/\s>]+)\s*\/>/) {
	    $text = $';
	    $name = $1;
	    $type = 'empty-element';
	
	# container element
	} elsif( $text =~ /^\s*<([^\/\s>]+)([^<>]*)>/) {
	    $text = $';
	    $name = $1;
	    $type = 'element';
	    my $atts = $2;
	    $atts =~ s/^\s+//;
	    push( @children, &parse_atts( $atts )) if $atts;
	    
	    # process child nodes
	    while( $text !~ /^<\/$name\s*>/ ) {
		my $newnode;
		($newnode, $text) = &get_node( $text );
		push( @children, $newnode );
	    }

	    # remove end tag
	    if( $text =~ /^<\/$name\s*>/ ) {  
		$text = $';
	    } else {
		&parse_error( "end tag for element <$name>" );
	    }
	} else {
	    if( $text =~ /^\s*<\/(\S+)/ ) {
		&parse_error( "missing start tag: $1" );
	    } else {
		&parse_error( "reserved character (<) in text" );
	    }
	}

    } else {
	&parse_error( "unidentified text" );
    }

    # create node
    $node = &make_node( $type, $name, $data, @children );
    my $n;
    foreach $n (@children) {
	$n->{'parent'} = $node;
    }

    return( $node, $text );
}


# parse_atts
# ----------
#
sub parse_atts {
    my( $text ) = @_;
    my @nodes = ();
    while( $text ) {
	if( $text =~ /\s*([^\s=]+)\s*=\s*([\"][^\"]*[\"])/ ||
	    $text =~ /\s*([^\s=]+)\s*=\s*([\'][^\']*[\'])/) {
	    $text = $';
	    my( $name, $data ) = ($1, $2);
	    push( @nodes, &make_node( 'attribute', $name, $data ));
	} elsif( $text =~ /^\s+/ ) {
	    $text = $';
	} else {
	    &parse_error( "attribute syntax" );
	}
    }
    return @nodes;
}


# make_node
# ---------
#
sub make_node {
    my( $type, $name, $data, @children ) = @_;
    my %newnode = (
		   'type' => $type,
		   'name' => $name,
		   'data' => $data,
		   'children' => \@children
		   );
    return \%newnode;
}


# parse_error
# -----------
#
sub parse_error {
    my( $reason ) = @_;
    die( "Parse error: $reason.\n" );
}


# process_tree
# ------------
#
sub process_tree {
    my( $node ) = @_;

    # root
    if( $node->{'type'} eq 'root' ) {
	# recurse over the children to traverse the tree
	my $child;
	foreach $child (@{$node->{'children'}}) {
	    &process_tree( $child );
	}

    # element
    } elsif( $node->{'type'} =~ /element$/ ) {

	# move/delete elements if they're in bad places
	&restructure_elements( $node );

	# recurse over the children to traverse the tree
	my $child;
	foreach $child (@{$node->{'children'}}) {
	    &process_tree( $child );
	}
    }
}


# get_descendants
# ---------------
# Find all matches of a descendant's name in a subtree at a 
# specified depth.
#
sub get_descendants { 
    my( $node, $descendant_name, $descendant_type, $depth ) = @_;
    my @results = ();
    # if result found, add to results list
    if(( $descendant_name && $node->{'name'} eq $descendant_name ) ||
       ( $descendant_type && $node->{'type'} eq $descendant_type )) {
	push( @results, $node );
    }
    # recurse if possible
    if( $node->{'type'} eq 'element' &&
       ((defined( $depth ) && $depth > 0) || !defined( $depth ))) {
	my $child;
	foreach $child (@{$node->{'children'}}) {
	    if( defined( $depth )) {
		push( @results, 
		     &get_descendants( $child, $descendant_name, 
				      $descendant_type, $depth - 1 ));
	    } else {
		push( @results, 
		     &get_descendants( $child, $descendant_name, 
				      $descendant_type ));
	    }
	}
    }
    return @results;
}


# serialize_tree
# --------------
# Print an object tree as XML text
#
sub serialize_tree {
    my( $node ) = @_;
    my $buf = "";

    # root node
    if( $node->{'type'} eq 'root' ) {
	my $n;
	foreach $n (@{$node->{'children'}}) {
	    $buf .= &serialize_tree( $n );
	}

    # element and empty-element
    } elsif( $node->{'type'} eq 'element' ||
	     $node->{'type'} eq 'empty-element' ) {
	$buf .= "<" . $node->{'name'};
	my $n;
	foreach $n (@{$node->{'children'}}) {
	    if( $n->{'type'} eq 'attribute' ) {
		$buf .= &serialize_tree( $n );
	    }
	}
	$buf .= ">";
	if( $node->{'type'} eq 'element' ) {
	    foreach $n (@{$node->{'children'}}) {
		if( $n->{'type'} ne 'attribute' ) {
		    $buf .= &serialize_tree( $n );
		}
	    }
	    $buf .= "</" . $node->{'name'} . ">";
	} else {
	    $buf .= "/>";
	}

    # attribute
    } elsif( $node->{'type'} eq 'attribute' ) {
	$buf .= " " . $node->{'name'} . "=" . $node->{'data'};

    # comment
    } elsif( $node->{'type'} eq 'comment' ) {
	$buf .= "<!--" . $node->{'data'} . "-->";

    # declaration
    } elsif( $node->{'type'} eq 'declaration' ) {
	$buf .= "<!" . $node->{'name'} . $node->{'data'} . ">" .
	    &space_after_start( "decl:" . $node->{'name'} );

    # CDMS
    } elsif( $node->{'type'} eq 'CDMS' ) {
	$buf .= "<![CDATA[" . $node->{'data'} . "]]>";

    # PI
    } elsif( $node->{'type'} eq 'PI' ) {
	$buf .= "<?" . $node->{'name'} . " " . $node->{'data'} . "?>" . 
	    &space_after_start( "pi:" . $node->{'name'} );

    # text
    } elsif( $node->{'type'} eq 'text' ) {
	$buf .= $node->{'data'};
    }
    return $buf;
}


# output_text
# -----------
# Find the special processing instructions in the text that denote 
# file boundaries, chop up the file into those regions, then save
# them out to files again.
#
sub output_text {
    my( $file, $text ) = @_;
    $text = "<?xml-file startfile: $file ?>" . $text . 
	"<?xml-file endfile ?>\n";
    my @filestack = ($file);
    my %data = ();
    while( $text ) {
	if( $text =~ /<\?xml-file\s+([^\s\?]+)\s*([^\?>]*)\?>/){
	    $data{ $filestack[ $#filestack ]} .= $`;
	    my( $mode, $rest ) = ($1, $2);
	    $text = $';
	    if( $mode eq 'startfile:' && $rest =~ /\s*(\S+)/ ) {
		push( @filestack, $1 );
	    } elsif( $mode eq 'endfile' ) {
		pop( @filestack );
	    } else {
		die( "Error with xml-file PIs: $mode, $rest" );
	    }
	} else {
	    $data{ $filestack[ $#filestack ]} .= $text;
	    $text = "";
	}
    }
    foreach $file (sort keys %data) {
	print "updated file: $file\n";
	if( open( F, ">$file" )) {
	    print F $data{ $file };
	    close F;
	} else {
	    print STDOUT "Warning: can't write to \"$file\"\n";
	}
    }
}

# restructure_elements
# --------------------
#
sub restructure_elements {
    my( $node ) = @_;
    # unwrap elements
    if( defined( $node->{'name'} ) &&
        defined( $unwrap_element{ $node->{'name'} })) {
	my $elem;
	foreach $elem (@{$unwrap_element{ $node->{'name'} }}) {
	    my $n;
	    foreach $n (&get_descendants( $node, $elem )) {
		&unwrap_element( $n ) if( $n->{'name'} eq $elem );
	    }
	}
    }
    # raise elements
    if( defined( $node->{'name'} ) &&
        defined( $raise_in_place{ $node->{'name'} })) {
	my $elem;
	foreach $elem (@{$raise_in_place{ $node->{'name'} }}) {
	    my $n;
	    foreach $n (&get_descendants( $node, $elem, undef, 1 )) {
		&raise_in_place( $n ) if( $n->{'name'} eq $elem );
	    }
	}
    }
    # raise elements and move backward
    if( defined( $node->{'name'} ) &&
        defined( $raise_and_move_backward{ $node->{'name'} })) {
	my $elem;
	foreach $elem (@{$raise_and_move_backward{ $node->{'name'} }}) {
	    my $n;
	    foreach $n (&get_descendants( $node, $elem, undef, 1 )) {
		&raise_and_move_backward( $n ) if( $n->{'name'} eq $elem );
	    }
	}
    }
}

# unwrap_element
# --------------
# delete an element, leaving its children (minus attributes) in its place
#
sub unwrap_element {
    my( $node ) = @_;
    # get list of children to save
    my @children_to_save = ();
    my $child;
    foreach $child (@{$node->{'children'}}) {
	push( @children_to_save, $child ) 
	    if( $child->{'type'} ne 'attribute' );
    }
    # delete node from parent's list
    my $count = &count_older_siblings( $node );
    # insert children in its place
    my $parent = $node->{'parent'};
    while( $child = pop @children_to_save ) {
	&insert_node( $child, $parent, $count );
    }
    # lose the node
    &delete_node( $node );
}

# count_older_siblings
# --------------------
# count the number of nodes that come before the selected node in the 
# attribute {'children'}
#
sub count_older_siblings {
    my( $node ) = @_;
    my $n;
    my $count = 0;
    foreach $n (@{$node->{'parent'}->{'children'}}) {
	last if( $n eq $node );
	$count ++;
    }	
    return $count;
}

# delete_node
# -----------
# remove a node from the hierarchy by erasing its parent's 
# reference to it, and its reference to its parents
#
sub delete_node {
    my( $node ) = @_;
    # if node has a parent...
    if( defined( $node->{'parent'} )) { 
	my $parent = $node->{'parent'};
	my @siblings = ();
	my $sib;
	# get list of siblings (not including the node to delete)
	foreach $sib (@{$parent->{'children'}}) {
	    push( @siblings, $sib ) unless( $node eq $sib );
	}
	# assign siblings back to parent, sans node
	$parent->{'children'} = \@siblings;
	# sever familial ties
	undef( $node->{'parent'} );
    }
}

# insert_node
# -----------
# place a node in a selected location, specified by its new 
# parent node and the number of older siblings it will have 
# (i.e., its index number in the array $parent->{'children'})
#
sub insert_node {
    my ( $new_node, $parent, $pos ) = @_; # $n = number of older siblings
    # sever old ties if necessary
    if( defined( $new_node->{'parent'} )) {
	&delete_node( $new_node );
    }
    # create new list of children
    my @siblings = ();
    my $n;
    my $count = 0;
    foreach $n (@{$parent->{'children'}}) {
	if( $count == $pos ) {
	    push( @siblings, $new_node );
	}
	push( @siblings, $n );
	$count++;
    }
    # assign children to parent
    $parent->{'children'} = \@siblings;
    # assign parent to new node
    $new_node->{'parent'} = $parent;
}

# raise_and_move_backward
# -----------------------
# raise a child to sibling level, directly before the parent
#
sub raise_and_move_backward {
    my( $node ) = @_;
    if( defined( $node->{'parent'} )) {
	my $parent = $node->{'parent'};
	my $count = &count_older_siblings( $parent );
	&insert_node( $node, $parent->{'parent'}, $count );
    }
}

# raise_and_move_forward
# ----------------------
# raise a child to sibling level, directly after the parent
#
sub raise_and_move_forward {
    my( $node ) = @_;
    if( defined( $node->{'parent'} )) {
	my $count = &count_older_siblings( $node->{'parent'} );
	&insert_node( $node, $node->{'parent'}->{'parent'}, $count +1 );
    }
}

# raise_in_place
# --------------
# split an element around a child and raise it to same level
#
# before: <a>xxx<b>yyy</b>zzz</a>
# after:  <a>xxx</a><b>yyy</b><a>zzz</a>
#
sub raise_in_place {
    my( $node ) = @_;
    if( defined( $node->{'parent'} ) &&
        defined( $node->{'parent'}->{'parent'} )) {
	# get lists of older and younger siblings
	my $parent = $node->{'parent'};
	my $n;
	my @older_sibs = ();
	my @younger_sibs = ();
	my $nodeseen = 0;
	foreach $n (@{$parent->{'children'}}) {
	    if( $node eq $n ) {
		$nodeseen = 1;
	    } else {
		push( @older_sibs, $n ) unless( $nodeseen );
		push( @younger_sibs, $n ) if( $nodeseen );
	    }
	}
	# reassign children for old parent
	$parent->{'children'} = \@older_sibs;
	# delete and insert node just after parent
	&delete_node( $node );
	my $count = &count_older_siblings( $parent );
	&insert_node( $node, $parent->{'parent'}, $count+1, );
	# make a new node to hold the younger siblings
	my $new_node = 
	    &make_node( 'element', $parent->{'name'}, undef, 
                        @younger_sibs );
	&insert_node( $new_node, $parent->{'parent'}, $count+2, );
	foreach $n (@{$new_node->{'children'}}) {
	    $n->{'parent'} = $new_node;
	}
    }
}
